/*

Digital Showroom               ▄ ▄ ▄
for Synchronet                 █████
Version 0.250413               ▐▄█▄▌ cf
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
by Craig Hendricks
codefenix@conchaos.synchro.net

ConstructiveChaos BBS:
  https://conchaos.synchro.net
 telnet://conchaos.synchro.net
    ssh://conchaos.synchro.net

*/

load("sbbsdefs.js");
require("dd_lightbar_menu.js", "DDLightbarMenu");
var sauce = load({},"sauce_lib.js");

const EXEC_PATH = backslash(js.exec_dir);
const SCROLLER_PATH = "../xtrn/scroller/scroller.js";
const SCROLLER_CMD_FMT = '?%s "%s" "%s" %s';
const TEMP_PATH = EXEC_PATH + "temp" + bbs.node_num;
const WIDTH = console.screen_columns;
const HEIGHT = console.screen_rows;
const HDR_FILE = "dsr" + (WIDTH >= 132 ? "_wide": "") + ".msg";
const WIDTH_MULTIPLIER = 8;
const HEIGHT_MULTIPLIER = 14;

var pathToMagick;
var resize;
var resizeMaxWidth = WIDTH * WIDTH_MULTIPLIER;
var resizeMaxHeight = HEIGHT * HEIGHT_MULTIPLIER;
var imgFormats;
var txtFormats;
var artFormats;
var arcFormats;
var arcCmdLins;

function printImVersionInfo () {
    if (file_exists(pathToMagick)) {
        log(pathToMagick + ' -version > "' + EXEC_PATH + 'info/About_ImageMagick.txt"');
        system.exec(pathToMagick + ' -version > "' + EXEC_PATH + 'info/About_ImageMagick.txt"');
    }
}

function showSixels(sixelPath) { 
    var image = new File(sixelPath);
    if (image.exists) {
        if (image.open("rb", true)) {            
            console.write(image.read());
            print(""); // prevents image destruction in IcyTerm
            image.close();
        }
    }
}

function handleImage(imgPath) {
    log(LOG_INFO, "viewing: '" + imgPath + "'");

    if (file_isdir(imgPath)) {
        var sxl_frames = directory(backslash(imgPath) + "*.sixel");
        if (sxl_frames.length > 1) {
            console.clear(false);
            for (var f = 0; f < sxl_frames.length; f++) {
                console.home(false);
                showSixels(sxl_frames[f]);
            }
        } else if (sxl_frames.length === 1) {
            console.clear(false);
            console.home(false);
            showSixels(sxl_frames[0]);
        }
    } else {
        if (file_exists(imgPath)) {
            console.clear(false);
            print("Preparing \x01b\x01h" + file_getname(imgPath) + "\x01w\x01h. Please wait.\x01n.\x01k\x01h..\x01n.\x01k\x01h.\x01w\x01h!");
            var tmpSixel = backslash(TEMP_PATH) + "temp" + bbs.node_num + "-%05d.sixel";
            var cmd = pathToMagick + " \"" + imgPath + "\" -coalesce " + (resize ? ("-resize " + resizeMaxWidth + "x" + resizeMaxHeight + " ") : "") + tmpSixel;
            var rslt = system.exec(cmd);
            var frames = directory(backslash(TEMP_PATH) + "temp" + bbs.node_num + "*.sixel");
            if (frames.length > 1) {
                print("\r\n\x01nThis image has \x01w\x01hANIMATION! \x01nLoop how many times? \x01k\x01h(max \x01n50\x01k\x01h)\x01n:");
                var loops = console.getnum(50);
                if (loops <= 0) {
                    loops = 1;
                }
                print("\r\n\x01nAnimation delay \x01k\x01h(in milliseconds; default \x01w\x01h15\x01k\x01h):");
                var frame_delay = console.getnum(50);
                if (frame_delay <= 0) {
                    frame_delay = 15;
                }
                console.clear(false);
                for (var l = 0; l < loops; l++) {
                    for (var f = 0; f < frames.length; f++) {
                        console.home(false);
                        showSixels(frames[f]);
                        if (l >= loops-1) { // cleanup on the last loop.
                            file_remove(frames[f]);
                        }
                        mswait(frame_delay);
                    }
                }
            } else if (frames.length === 1) {
                console.clear(false);
                console.home(false);
                showSixels(frames[0]);
                file_remove(frames[0]);
            } else {
                print("failed.");
                log(LOG_WARNING, "No output found for '" + imgPath + "'.");
            }
            if (rslt !== 0) {
                log(LOG_WARNING, "Convert for '" + imgPath + "' rslt: " + rslt);
            }
        }
    }
}

function unarchiveFile(archiveFile, destPath) {
    mkdir(destPath);
    var extracted = false;
    for (var f in arcFormats) {
        if (file_getext(archiveFile).toUpperCase()==arcFormats[f].toUpperCase() && arcCmdLins[f]) {
            var unzipcmd;
            unzipcmd = bbs.cmdstr( arcCmdLins[f].replace("~path~", '"' + destPath + '"').replace("~file~", '"' + archiveFile + '"') );
            extracted = (system.exec( unzipcmd ) === 0);            
            break;
        }        
    }
    
    // The built-in Archive class is problematic and fails to successfully extract 
    // anything from lots of fairly typical ZIP files, so we're going to use it as 
    // a last resort in case the above extraction attempt fails.
    if (!extracted) {
        log("using Archive class for: '" + archiveFile + "'.");
        var a = new Archive(archiveFile);
        try {                
            a.extract(destPath);   
        } catch (e) {
            log(LOG_ERROR, e);
        }
    }
}

function deleteTempDir(path) {
    // This *should* dynamically clean up a directory and its contents, but it doesn't...
    /*var df = directory(backslash(path) + "*");
    for (var i = 0; i < df.length; i++) {
        if (file_isdir(df[i])) {
            deleteTempDir(df[i]);
        } else {
            file_removecase(df[i]); // doesn't delete read-only files..
            if (file_exists(df[i])) { 
                system.exec("del \"" + df[i] + "\"");
                if (file_exists(df[i])) { 
                    log("Tougher than it looks...");
                } else {
                    log("Got 'em, sir.");
                }
            }
        }
    }
    rmdir(path); // This *should* remove the path and all contents, but it doesn't...  */
    
    // This makes a system call to handle the directory & file cleanup.. lifted from DD_Ar_Viewer
    path = path.replace(/\\$/, "").replace(/\/$/, ""); // trim off the right trailing \ or / if there is one...
	if ( /^WIN/i.test(system.platform) ) {
		system.exec("RD \"" + (path) + "\" /s /q");
	} else {
		system.exec("rm -rf \"" + (path) + "\"");
    }
}

function dir_getname(path) {
    var dirname = "";
    var levels = path.replace(/\\/g, "/").split(/\//);
    if (levels[levels.length-1]==="") {
        dirname = levels[levels.length-2];        
    } else {
        dirname = levels[levels.length-1];
    }
    return dirname;
}

function browseFiles(path, parentPath) {
    var func_exit = false;
    var selectedFile = "";
    var last_selected = ""
    var last_selected_index;
    var dir_contents = directory(backslash(path) + "*");
    var option_index = 0;
    var ext;
    var fname;
    dir_contents.sort(function (a, b) {
        return a.toLowerCase().localeCompare(b.toLowerCase());
    });

    while (bbs.online && !js.terminated && !func_exit) {
        console.clear();
        console.printfile(EXEC_PATH + HDR_FILE, P_NOABORT);
        option_index = 0;
        console.gotoxy(35, HEIGHT-1);
        console.center("\x01w\x01hQ\x01k\x01\h) \x01n Go back");
        var lbMenu = new DDLightbarMenu(17, 10, WIDTH-33, HEIGHT-11);
        for (var d = 0; d < dir_contents.length; d++) {
            if (!file_isdir(dir_contents[d])) {
                if (dir_contents[d].indexOf(".")>=0) { // TODO: include files without extensions
                    ext = (file_getext(dir_contents[d])+"").toUpperCase();
                    if ( imgFormats.indexOf(ext) >= 0 ||
                         txtFormats.indexOf(ext) >= 0 ||
                         artFormats.indexOf(ext) >= 0 ||
                         arcFormats.indexOf(ext) >= 0 ) {
                        fname = file_getname(dir_contents[d]);
                        lbMenu.Add( fname, dir_contents[d] );
                        if (last_selected===dir_contents[d]) {
                            last_selected_index = option_index;
                        }
                        option_index = option_index + 1;
                    }
                }
            } else if ( file_isdir(dir_contents[d]) ) {
                fname = dir_getname(dir_contents[d]);
                lbMenu.Add( fname.toUpperCase(), dir_contents[d] );
                if (last_selected===dir_contents[d]) {
                    last_selected_index = option_index;
                }
                option_index = option_index + 1;
            }
        }
        console.gotoxy(1, 8);
        console.center("\x01nBrowsing \x01w\x01h" + (parentPath ? (parentPath + "/") : "").replace(/\//g, "\x01k/\x01w") + dir_getname(path) + "\x01n \x01k\x01h(\x01w\x01h" + option_index + " \x01nfile"+(option_index> 1?"s":"")+"\x01k\x01h)");
        if (option_index == 0) {
            console.center("\x01mNo applicable files!");
            console.gotoxy((WIDTH/2)-8, HEIGHT/2);
            console.pause();
            return;
        }
        console.center("\x01nUse \x01w\x01h" + ascii(24) + " \x01nand \x01w\x01h" + ascii(25) +" \x01nkeys to scroll, \x01w\x01hENTER \x01nselects");
        lbMenu.colors.itemColor = "\x01k\x01h";
        lbMenu.colors.selectedItemColor = "\x01w\x01h\x01" + "4";
        lbMenu.AddAdditionalQuitKeys("qQ");
        lbMenu.borderEnabled = true;
        lbMenu.scrollbarEnabled = true;
        lbMenu.ampersandHotkeysInItems = false;
        if (last_selected) {
            lbMenu.SetSelectedItemIdx(last_selected_index);
        }
        selectedFile = lbMenu.GetVal();

        if (!selectedFile) {
            func_exit = true;
        } else {
            last_selected_index = 0;
            last_selected = selectedFile;

            print("\x01n\x010\x01L");
            ext = file_getext(selectedFile);
            fname = file_getname(selectedFile);

            if ( file_isdir(selectedFile) ) {
                browseFiles(selectedFile, (parentPath ? (parentPath + "/") : "") + dir_getname(path));
            }
            else if (typeof ext != "undefined" && arcFormats.indexOf(ext.toUpperCase()) >= 0) {
                var destDir = backslash(TEMP_PATH) + fname;
                if (!file_isdir(destDir)) {
                    unarchiveFile(selectedFile, destDir);                    
                }
                if (file_isdir(destDir)) {
                    browseFiles(destDir, (parentPath ? (parentPath + "/") : "") + dir_getname(path));
                    deleteTempDir(destDir);
                }
            } else if (txtFormats.indexOf(ext.toUpperCase()) >= 0 || artFormats.indexOf(ext.toUpperCase()) >= 0) {
                if (file_exists(SCROLLER_PATH)) {
                    bbs.exec( format(SCROLLER_CMD_FMT, SCROLLER_PATH, selectedFile, fname, "top" ), 0, EXEC_PATH );
                } else {
                    var sdat = new sauce.read(selectedFile);
                    console.clear(false);
                    if (sdat.ice_color || ext.toUpperCase()===".ICE" ) {
                        print("\x1b[?33;35h"); // switch iCE colors on
                    }
                    print("\x1b[;11*r"); // throttle speed down to 115200
                    console.printfile(selectedFile, P_NOPAUSE | P_CPM_EOF | P_NOATCODES);
                    print("\x1b[*r"); // reset speed
                    print("\x1b[?33;35l"); // switch iCE colors off
                    console.pause();                    
                }                               
            } else if (imgFormats.indexOf(ext.toUpperCase()) >= 0) {
                handleImage(selectedFile);
                console.gotoxy(1, HEIGHT-1);
                console.pause();
            }
        }
    }

    printf("\x01n\x010\x01L");

    console.clear();
    console.home();
}

function mainMenu() {
    var func_exit = false;
    var selectedOption = "";
    var last_selected;
    var jsonPaths = "";
    var fPaths = new File(EXEC_PATH + "paths.json");
    if (fPaths.open("r")) {
        jsonPaths = JSON.parse(fPaths.read());
        jsonPaths.push({"name":"- INFO -", "path": EXEC_PATH + "info", "resize":false});
        fPaths.close();
        while (bbs.online && !js.terminated && !func_exit) {
            printf("\x01n\x010\x01L");
            console.home();
            console.printfile(EXEC_PATH + HDR_FILE, P_NOABORT);
            console.gotoxy(35, HEIGHT-1);
            console.center("\x01w\x01hQ\x01k\x01\h) \x01w\x01hQ\x01nuit");
            var lbMenu = new DDLightbarMenu(18, 9, WIDTH - 35, HEIGHT - 10);
            for (var s = 0; s < Object.keys(jsonPaths).length; s++) {
                lbMenu.Add( jsonPaths[s].name, s+1 );
            }
            console.gotoxy(20, 8);
            console.center("\x01nUse \x01w\x01h" + ascii(24) + " \x01nand \x01w\x01h" + ascii(25) +" \x01nkeys to scroll, \x01w\x01hENTER \x01nselects");
            lbMenu.colors.itemColor = "\x01k\x01h";
            lbMenu.colors.selectedItemColor = "\x01w\x01h\x01" + "4";
            lbMenu.AddAdditionalQuitKeys("qQ");
            lbMenu.borderEnabled = true;
            lbMenu.scrollbarEnabled = true;
            if (last_selected) {
                lbMenu.SetSelectedItemIdx(last_selected);
            }
            selectedOption = lbMenu.GetVal();

            if (!selectedOption) {
                func_exit = true;
            } else {
                selectedOption = parseInt(selectedOption) - 1;
                last_selected = selectedOption;
                resize = jsonPaths[selectedOption].resize;
                browseFiles(jsonPaths[selectedOption].path);
            }
        }
    }
}

function init() {
    if (!file_isdir(TEMP_PATH)) {
        mkdir(TEMP_PATH);
    }
    var fIni = new File(EXEC_PATH + 'settings.ini');
    fIni.open('r');
    const settings = {
        root: fIni.iniGetObject()
    };
    fIni.close();
    fIni = undefined;
    pathToMagick = settings.root.pathToMagick;
    resize = settings.root.resize;
    imgFormats = settings.root.imgFormats.split(';');
    txtFormats = settings.root.txtFormats.split(';');
    artFormats = settings.root.artFormats.split(';');
    arcFormats = settings.root.arcFormats.split(';');
    arcCmdLins = settings.root.arcCmdLins.split(';');
    printImVersionInfo();
    
    if (!file_exists(pathToMagick)) {
        log(LOG_ERROR, "ImageMagick not found: " + pathToMagick);
        exit();
    }        
        
    if (settings.root.unsupported_sixel_terminal_warning && console.cterm_version < 1189) {
        print("\x01k\x01hThis program makes use of \x01w\x01hSIXEL \x01k\x01hgraphics.\r\n");
        print("\x01k\x01hFor best results, come back using \x01w\x01hSyncTERM \x01k\x01hor \x01w\x01hMagiTerm\x01k\x01h.\r\n");
        print("You may see mixed (or no) results in the terminal you're currently using.\r\n");
        if (console.noyes("Run anyway")) {
            exit();
        }    
    }    
}

var filename = argv[0];

init();
if (filename) { // === "" || filename === undefined) {
    handleImage(filename);
} else {
    mainMenu();
}
